from relayserver.real.ireal_server import IRealServer


class EchoServer(IRealServer):
    def __init__(self, connection):
        self.__connection = connection
        
    def receive_data(self, proxied_data):
        self.__connection.transport.write(proxied_data)

    def shutdown(self):
        pass



from threading import RLock
from struct import unpack


class ReadBuffer(object):
    """This class receives incoming data, and stores it in a list of extents 
    (also referred to as buffers). It allows us to leisurely pop off sequences 
    of bytes, which we build from the unconsumed extents. As the extents are 
    depleted, we maintain an index to the first available, non-empty extent. 
    We will only occasionally cleanup.
    """

    __locker = RLock()
# TODO: Reduce this for testing.
    __cleanup_interval = 100
    
    def __init__(self):
        self.__buffers = []
        self.__length = 0
        self.__read_buffer_index = 0
        self.__hits = 0

    def push(self, data):
        with self.__class__.__locker:
            self.__buffers.append(data)
            self.__length += len(data)

    def read_message(self):
        """Try to read a message from the buffered data. A message is defined 
        as a 32-bit integer size, followed that number of bytes. First we try
        to non-destructively read the integer. Then, we try to non-
        destructively read the remaining bytes. If both are successful, we then
        go back to remove the span from the front of the buffers.
        """

        with self.__class__.__locker:
            result = self.__passive_read(4)
            if result is None:
                return None

            (four_bytes, last_buffer_index, updates1) = result
            (length,) = unpack('>I', four_bytes)

            result = self.__passive_read(length, last_buffer_index)
            if result is None:
                return None

            (data, last_buffer_index, updates2) = result

            # If we get here, we found a message. Remove it from the buffers.

            for updates in (updates1, updates2):
                for update in updates:
                    (buffer_index, buffer_, length_consumed) = update
                    self.__buffers[buffer_index] = buffer_ if buffer_ else ''
                    self.__length -= length_consumed

            self.__read_buffer_index = last_buffer_index
            self.__hits += 1
        
            if self.__hits >= self.__class__.__cleanup_interval:
                self.cleanup()
                self.__hits = 0

        return data

    def __passive_read(self, length, start_buffer_index=None):
        """Read the given length of bytes, or return None if we can't provide 
        [all of] them yet. When the given length is available but ends in the 
        middle of a buffer, we'll split the buffer. We do this to make it 
        simpler to continue from that point next time (it's always simpler to 
        start at the beginning of a buffer), as well as simpler to remove the
        found bytes later, if need be.
        """

        if length > self.__length:
            return None
    
        with self.__class__.__locker:
            collected = []
            need_bytes = length
            i = start_buffer_index if start_buffer_index is not None \
                                   else self.__read_buffer_index

            updates = []
            while need_bytes > 0:
                len_current_buffer = len(self.__buffers[i])
                
                if need_bytes >= len_current_buffer:
                    # We need at least as many bytes as are in the current 
                    # buffer. Consume them all.
                
                    collected.append(self.__buffers[i][:])
                    updates.append((i, [], len_current_buffer))
                    need_bytes -= len_current_buffer
                else:
                    # We need less bytes than are in the current buffer. Slice
                    # the current buffer in half, even if the data isn't going
                    # anywhere [yet].
                
                    first_half = self.__buffers[i][:need_bytes]
                    second_half = self.__buffers[i][need_bytes:]

                    self.__buffers[i] = first_half
                    self.__buffers.insert(i + 1, second_half)
                    
                    # We only mark the buffer that came from the first half as
                    # having an update (the second half of the buffer wasn't 
                    # touched).
                    collected.append(first_half)
                    updates.append((i, [], need_bytes))
                    need_bytes = 0

                i += 1

        sequence = ''.join(collected)

        return (sequence, i, updates)
    
    def cleanup(self):
        """Clip buffers that the top of our list that have been completely 
        exhausted.
        """
# TODO: Test this.
        with self.__class__.__locker:

            while self.__read_buffer_index > 0:
                del self.__buffers[0]
                self.__read_buffer_index -= 1





# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

import sys

from argparse import ArgumentParser
from threading import Lock
from collections import OrderedDict
from twisted.internet import reactor, protocol
from twisted.python.log import startLogging
from twisted.python import log

from relayserver.read_buffer import ReadBuffer
from relayserver.message_types.hello_pb2 import Hello
from relayserver.message_types import build_msg_cmd_connopen,\
                                      build_msg_cmd_conndrop,\
                                      build_msg_data_hphelloresponse

from relayserver.base_protocol import BaseProtocol

ports = None

# TODO: As Twisted mostly runs synchronously, see if there's someway we can 
#       spin-off write requests so that control can return to the reactor. 


class _AssignmentManager(object):
    # Lists of current connections.
    __client_list = OrderedDict()
    __hp_assigned_list = OrderedDict()
    __hp_waiting_list = OrderedDict()

    # Forward and reverse maps expressing assignments.
    __map_client_to_hp = {}
    __map_hp_to_client = {}

    __locker = Lock()

    def queue_new_hp(self, hp_connection):
        cls = self.__class__

        with cls.__locker:
            cls.__hp_waiting_list[hp_connection.session_id] = hp_connection

    def get_assigned_hp(self, client_session_id):
        cls = self.__class__
        
        with cls.__locker:
            return cls.__map_client_to_hp[client_session_id] \
                    if client_session_id \
                    in cls.__map_client_to_hp \
                    else None

    def get_client_for_hp(self, hp_session_id):
        cls = self.__class__
        
        with cls.__locker:
            return cls.__map_hp_to_client[hp_session_id] \
                    if hp_session_id \
                    in cls.__map_hp_to_client \
                    else None

    def assign_new_client(self, client_connection):
        cls = self.__class__
        
        with cls.__locker:
            # Make sure the host-process is listening on the command-channel.
            if CommandServer.get_command_channel() is None:
                log.msg("We're denying new client with session-ID (%d) "
                        "because a host-process command-channel is not "
                        "connected (assuming no HP available)." % 
                        (client_connection.session_id))
                
                return False
    
            # Make sure there is at least one host-process connection waiting
            # to service a request.

            if not cls.__hp_waiting_list:
                log.msg("We're denying new client with session-ID (%d) "
                        "because there are no available host-processes." % 
                        (client_connection.session_id))
                
                return False

            # Dequeue an unassigned HP connection.
            for available_hp_connection in cls.__hp_waiting_list.itervalues():
                hp_connection = available_hp_connection
                break
            
            del cls.__hp_waiting_list[hp_connection.session_id]

            # Assign the HP connection.
            cls.__map_client_to_hp[client_connection.session_id] = \
                hp_connection
            cls.__map_hp_to_client[hp_connection.session_id] = \
                client_connection

            # Store the assigned HP connection.
            cls.__hp_assigned_list[hp_connection.session_id] = \
                hp_connection

            # Store the client connection.
            cls.__client_list[client_connection.session_id] = self

            log.msg("Client with session-ID (%d) has been assigned to host-"
                    "process with session-ID (%d)." % 
                    (client_connection.session_id, hp_connection.session_id))

        # Now, announce the assignment on the command-channel, and respond 
        # to the client. Note that, although we don't guarantee the order,
        # the client would first have to receive the message and -then-
        # send data, whereas the host process may just have to perform a
        # couple of trivial operations, if any, to prepare a session that
        # should already be ready to go.

        # Emit an assignment message on the command-channel.
        CommandServer.get_command_channel().announce_assignment(hp_connection)

        return True

    def connection_lost_from_client(self, session_id):
        cls = self.__class__
        
        with cls.__locker:
            if session_id in cls.__map_client_to_hp:
                # A client connection dropped. This connection will only exist 
                # under assignment.
    
                mapped_hp = cls.__map_client_to_hp[session_id]
                
                log.msg("Client with session-ID (%d) mapped to host-process "
                        "with session-ID (%d) has dropped." % 
                        (session_id, mapped_hp.session_id))
    
                del cls.__map_client_to_hp[session_id]
                del cls.__map_hp_to_client[mapped_hp.session_id]
                del cls.__client_list[session_id]
                del cls.__hp_assigned_list[mapped_hp.session_id]
    
                command_channel = CommandServer.get_command_channel()
                if command_channel is not None:
                    command_channel.announce_drop(mapped_hp)
    
                reactor.callLater(5, mapped_hp.transport.loseConnection)

    def connection_lost_from_hp(self, session_id):
        cls = self.__class__
        
        with cls.__locker:
            if session_id in cls.__hp_waiting_list:
                log.msg("Unassigned host-process with session-ID (%d) has "
                        "dropped." % (session_id))
    
                del cls.__hp_waiting_list[session_id]
                
            elif session_id in cls.__map_hp_to_client:
                # An HP connection dropped. We've stated elsewhere that if this
                # is currently assigned, the result is undefined. We do our 
                # honest and simplest best, here, though.
    
                mapped_client = cls.__map_hp_to_client[session_id]
    
                log.msg("Host-process with session-ID (%d) mapped to client "
                        "with session-ID (%d) has dropped." % 
                        (session_id, mapped_client.session_id))
    
                del cls.__map_client_to_hp[mapped_client.session_id]
                del cls.__map_hp_to_client[session_id]
                del cls.__client_list[mapped_client.session_id]
                del cls.__hp_assigned_list[session_id]
    
                mapped_client.transport.loseConnection()

_assignments = _AssignmentManager()


class TrivialClientServer(BaseProtocol):
    """Handles operations for the client data channel."""

    def connectionMade(self):
        log.msg("Client with session-ID (%d) has connected." % 
                (self.session_id))

        try:
            if _assignments.assign_new_client(self) is False:
                self.transport.loseConnection()
        except:
            log.err()

    def dataReceived(self, data):
        try:
            hp_connection = _assignments.get_assigned_hp(self.session_id)
            hp_connection.transport.write(data)
        except:
            log.err()

    def connectionLost(self, reason):
        try:
            _assignments.connection_lost_from_client(self.session_id)
        except:
            log.err()

    @property
    def session_id(self):
        return self.transport.sessionno
    

class HostProcessServer(BaseProtocol):
    """Handles operations for the host-process data channel."""

    def __init__(self):
        self.__buffer = ReadBuffer()

    def dataReceived(self, data):
        cls = self.__class__

        try:
            client = _assignments.get_client_for_hp(self.session_id)
            if client is not None:
                client.transport.write(data)
                return
        
            self.__buffer.push(data)
    
            message_raw = self.__buffer.read_message()
            if message_raw is None:
                return
    
            hello = self.parse_or_raise(message_raw, Hello)
            self.__handle_hostprocess_hello()
        except:
            log.err()

    def connectionMade(self):
        log.msg("A data connection has been established with connection "
                "having session-ID (%d), but it will need to say hello." % 
                (self.session_id))

    def connectionLost(self, reason):
        try:
            _assignments.connection_lost_from_hp(self.session_id)
        except:
            log.err()
        
    def __handle_hostprocess_hello(self):
        """Handle a host-process hello."""

        cls = self.__class__

        try:
            log.msg("Host-process with session-ID (%d) has said hello." % 
                    (self.session_id))
    
            _assignments.queue_new_hp(self)
    
            host_info = self.transport.getHost()
            response = build_msg_data_hphelloresponse(self.session_id, 
                                                      host_info.host, 
                                                      ports[2])
    
            self.write_message(response)
        except:
            log.err()

    @property
    def session_id(self):
        return self.transport.sessionno
    

class CommandServer(BaseProtocol):
    """Handles operations for host-process command-channel."""

    __command_channel = None

    def __init__(self):
        self.__buffer = ReadBuffer()

    def dataReceived(self, data):
        self.__buffer.push(data)

        message_raw = self.__buffer.read_message()
        if message_raw is None:
            return

        # We don't actually expect any data to be sent to us. Do not take any 
        # further action, here.

    def connectionLost(self, reason):
        log.msg("Command channel dropped.")
        self.__class__.__command_channel = None
        
    def connectionMade(self):
        log.msg("Command channel connected.")
        self.__class__.__command_channel = self
        
    @classmethod
    def get_command_channel(cls):
        return cls.__command_channel

    def announce_assignment(self, hp_connection):
        log.msg("Announcing assignment of new client to HP session "
                      "(%d)." % (hp_connection.session_id))
        
        command = build_msg_cmd_connopen(hp_connection.session_id)
        self.write_message(command)
                
    def announce_drop(self, hp_connection):
        log.msg("Announcing drop of client assigned to HP session "
                      "(%d)." % (hp_connection.session_id))

        command = build_msg_cmd_conndrop(hp_connection.session_id)
        self.write_message(command)


class _GeneralFactory(protocol.ServerFactory):
    def __init__(self, description):
        """ServerFactory doesn't have a constructor and it's not a new-style 
        class, so we can't call it."""
        
        self.__description = description

    def __repr__(self):
        return ("GeneralFactory(%s)" % (self.__description))

def start_relay(dport, cport, tport):
    global ports

    ports = (dport, cport, tport)

    #startLogging(sys.stdout)

    relayFactory = _GeneralFactory("HostProcessServer")
    relayFactory.protocol = HostProcessServer
    relayFactory.connections = {}

    commandFactory = _GeneralFactory("CommandServer")
    commandFactory.protocol = CommandServer
    commandFactory.connections = {}

    trivialFactory = _GeneralFactory("TrivialClientServer")
    trivialFactory.protocol = TrivialClientServer
    trivialFactory.connections = {}

    reactor.listenTCP(ports[0], relayFactory)
    reactor.listenTCP(ports[1], commandFactory)
    reactor.listenTCP(ports[2], trivialFactory)
    reactor.run()




# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)



DESCRIPTOR = descriptor.FileDescriptor(
  name='hello.proto',
  package='relay',
  serialized_pb='\n\x0bhello.proto\x12\x05relay\"\x18\n\x05Hello\x12\x0f\n\x07version\x18\x01 \x02(\x05\"V\n\x18HostProcessHelloResponse\x12\x12\n\nsession_id\x18\x01 \x02(\x05\x12\x12\n\nrelay_host\x18\x02 \x02(\t\x12\x12\n\nrelay_port\x18\x03 \x02(\t')




_HELLO = descriptor.Descriptor(
  name='Hello',
  full_name='relay.Hello',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='version', full_name='relay.Hello.version', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=22,
  serialized_end=46,
)


_HOSTPROCESSHELLORESPONSE = descriptor.Descriptor(
  name='HostProcessHelloResponse',
  full_name='relay.HostProcessHelloResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='session_id', full_name='relay.HostProcessHelloResponse.session_id', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='relay_host', full_name='relay.HostProcessHelloResponse.relay_host', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='relay_port', full_name='relay.HostProcessHelloResponse.relay_port', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=48,
  serialized_end=134,
)

DESCRIPTOR.message_types_by_name['Hello'] = _HELLO
DESCRIPTOR.message_types_by_name['HostProcessHelloResponse'] = _HOSTPROCESSHELLORESPONSE

class Hello(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _HELLO
  
  # @@protoc_insertion_point(class_scope:relay.Hello)

class HostProcessHelloResponse(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _HOSTPROCESSHELLORESPONSE
  
  # @@protoc_insertion_point(class_scope:relay.HostProcessHelloResponse)

# @@protoc_insertion_point(module_scope)



from struct import unpack
from threading import Lock

from relayserver.base_protocol import BaseProtocol


class EndpointBaseProtocol(BaseProtocol):
    __locker = Lock()

    def __init__(self):
        self.__buffer = ""
    
    def push_data(self, data):
        """Data has been received while we are still in an unconfigured state. 
        Push all data until we can see a complete message.
        """
        
        with self.__class__.__locker:
            self.__buffer += data
    
    def get_and_clear_buffer(self):
        """This is called every time we receive data, but it's only really of 
        use immediately following being configured, just in case we pushed any 
        data that came in right behind the initial message."""
        
        with self.__class__.__locker:
            buffer_ = self.__buffer
            self.__buffer = ""

        return buffer_
    
    def get_initial_message(self):
        """We expect exactly one message within the entire session. Return None 
        if not found.
        """

        length_length = 4
        current_bytes = len(self.__buffer)
        if current_bytes < length_length:
            return None

        length_bytes = self.__buffer[0:length_length]
        (length,) = unpack('>I', length_bytes)

        if current_bytes < (length_length + length):
            return None
            
        message = self.__buffer[length_length:(length_length + length)]
        self.__buffer = self.__buffer[(length_length + length):]

        return message



class IRealServer(object):
    """This receives the actual proxied data (going toward the server)."""
        
    def receive_data(self, proxied_data):
        raise NotImplementedError()

    def shutdown(self):
        raise NotImplementedError()



from argparse import ArgumentParser

from relayserver.main import start_relay

def main():
    parser = ArgumentParser(description='Start a protocol-agnostic relay '
                                        'server.')

    parser.add_argument('dport', 
                        nargs='?', 
                        default=8000, 
                        type=int, 
                        help="Port for host-process data-channel connections.")

    parser.add_argument('cport', 
                        nargs='?', 
                        default=8001, 
                        type=int, 
                        help="Port for host-process command-channel "
                             "connections.")

    parser.add_argument('tport', 
                        nargs='?', 
                        default=8002, 
                        type=int, 
                        help="Port for client data-channel connections.")

    args = parser.parse_args()

    start_relay(args.dport, args.cport, args.tport)

if __name__ == '__main__':
    main()



# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)



DESCRIPTOR = descriptor.FileDescriptor(
  name='command.proto',
  package='relay',
  serialized_pb='\n\rcommand.proto\x12\x05relay\"=\n\x1e\x43lientConnectionOpenProperties\x12\x1b\n\x13\x61ssigned_to_session\x18\x01 \x02(\x05\"4\n\x1e\x43lientConnectionDropProperties\x12\x12\n\nsession_id\x18\x01 \x02(\x05\"\x85\x02\n\x07\x43ommand\x12\x0f\n\x07version\x18\x01 \x02(\x05\x12\x30\n\x0cmessage_type\x18\x02 \x02(\x0e\x32\x1a.relay.Command.MessageType\x12>\n\x0fopen_properties\x18\x03 \x01(\x0b\x32%.relay.ClientConnectionOpenProperties\x12>\n\x0f\x64rop_properties\x18\x04 \x01(\x0b\x32%.relay.ClientConnectionDropProperties\"7\n\x0bMessageType\x12\x13\n\x0f\x43ONNECTION_OPEN\x10\x00\x12\x13\n\x0f\x43ONNECTION_DROP\x10\x01')



_COMMAND_MESSAGETYPE = descriptor.EnumDescriptor(
  name='MessageType',
  full_name='relay.Command.MessageType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    descriptor.EnumValueDescriptor(
      name='CONNECTION_OPEN', index=0, number=0,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='CONNECTION_DROP', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=348,
  serialized_end=403,
)


_CLIENTCONNECTIONOPENPROPERTIES = descriptor.Descriptor(
  name='ClientConnectionOpenProperties',
  full_name='relay.ClientConnectionOpenProperties',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='assigned_to_session', full_name='relay.ClientConnectionOpenProperties.assigned_to_session', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=24,
  serialized_end=85,
)


_CLIENTCONNECTIONDROPPROPERTIES = descriptor.Descriptor(
  name='ClientConnectionDropProperties',
  full_name='relay.ClientConnectionDropProperties',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='session_id', full_name='relay.ClientConnectionDropProperties.session_id', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=87,
  serialized_end=139,
)


_COMMAND = descriptor.Descriptor(
  name='Command',
  full_name='relay.Command',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='version', full_name='relay.Command.version', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='message_type', full_name='relay.Command.message_type', index=1,
      number=2, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='open_properties', full_name='relay.Command.open_properties', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='drop_properties', full_name='relay.Command.drop_properties', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _COMMAND_MESSAGETYPE,
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=142,
  serialized_end=403,
)

_COMMAND.fields_by_name['message_type'].enum_type = _COMMAND_MESSAGETYPE
_COMMAND.fields_by_name['open_properties'].message_type = _CLIENTCONNECTIONOPENPROPERTIES
_COMMAND.fields_by_name['drop_properties'].message_type = _CLIENTCONNECTIONDROPPROPERTIES
_COMMAND_MESSAGETYPE.containing_type = _COMMAND;
DESCRIPTOR.message_types_by_name['ClientConnectionOpenProperties'] = _CLIENTCONNECTIONOPENPROPERTIES
DESCRIPTOR.message_types_by_name['ClientConnectionDropProperties'] = _CLIENTCONNECTIONDROPPROPERTIES
DESCRIPTOR.message_types_by_name['Command'] = _COMMAND

class ClientConnectionOpenProperties(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _CLIENTCONNECTIONOPENPROPERTIES
  
  # @@protoc_insertion_point(class_scope:relay.ClientConnectionOpenProperties)

class ClientConnectionDropProperties(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _CLIENTCONNECTIONDROPPROPERTIES
  
  # @@protoc_insertion_point(class_scope:relay.ClientConnectionDropProperties)

class Command(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _COMMAND
  
  # @@protoc_insertion_point(class_scope:relay.Command)

# @@protoc_insertion_point(module_scope)



from struct import pack
from google.protobuf.message import DecodeError
from twisted.internet.protocol import Protocol
from twisted.python import log

from relayserver.utility import get_hex_dump

class BaseProtocol(Protocol):
    def write_message(self, message):
        data = message.SerializeToString()
        message_len = len(data)
        prefix = pack('>I', message_len)
        
        total_len = len(prefix) + message_len
    
        log.msg("Message type [%s] serialized into (%d) bytes and wrapped as "
                "(%d) bytes." % 
                (message.__class__.__name__, message_len, total_len))

        self.transport.write(prefix)
        self.transport.write(data)

    def parse_or_raise(self, message_raw, type_):
        log.msg("Parsing [%s] in (%d) bytes." % (type_.__name__, len(message_raw)))

        message = type_()

        try:
            message.ParseFromString(message_raw)
        except DecodeError:
            raise
        else:
            if message.IsInitialized() is False:
                raise Exception("Message parse with type [%s] resulted in "
                                "uninitialized message." % (type_))

        return message




#!/usr/bin/python

from argparse import ArgumentParser
from sys import stdout

from twisted.internet import reactor
from twisted.internet.protocol import ReconnectingClientFactory
from twisted.python.log import startLogging
from twisted.python import log

from relayserver.message_types.command_pb2 import Command
from relayserver.message_types.hello_pb2 import HostProcessHelloResponse
from relayserver.message_types import build_msg_data_hphello
from relayserver.base_protocol import BaseProtocol
from relayserver.endpoint import EndpointBaseProtocol
from relayserver.read_buffer import ReadBuffer

from relayserver.config import EndpointServer


class HostProcess(EndpointBaseProtocol):
    """This is a connection to the relay server. It knows how to configure 
    itself, and then forward all subsequent data to the "RealServer" instance.
    """
    
    def __init__(self, factory):
        EndpointBaseProtocol.__init__(self)

        self.__factory = factory
        self.__configured = False

        self.__session_id = None;
        self.__relay_host = None;
        self.__relay_port = None;

        self.__buffer_cleared = False
        self.__real_server = EndpointServer(self)
    
    def connectionLost(self, reason):
        log.msg("Host-process with session-no (%d) has had its connection "
                "dropped." % (self.__session_id))

        self.__real_server.shutdown()
    
    def connectionMade(self):
        log.msg("We've successfully connected to the relay server. Sending "
                "hello.")

        hello = build_msg_data_hphello()
        self.write_message(hello)

    def __handle_configuration_data(self, data):
        """Our connection has not been configured yet."""

        self.push_data(data)

        message_raw = self.get_initial_message()
        if message_raw is None:
            return

        response = self.parse_or_raise(message_raw, HostProcessHelloResponse)

        self.__session_id = response.session_id;
        self.__relay_host = response.relay_host;
        self.__relay_port = int(response.relay_port);

        log.msg("Received hello response: SESSION-NO=(%d) RHOST=[%s] "
                      "RPORT=(%d)" % 
                      (self.__session_id, self.__relay_host, 
                       self.__relay_port))

        self.__configured = True
        self.__factory.resetDelay()
    
    def dataReceived(self, data):
        try:
            if self.__configured is False:
                self.__handle_configuration_data(data)
            else:
                if self.__buffer_cleared is False:
                    data = self.get_and_clear_buffer() + data
                    self.__buffer_cleared = True

                self.__real_server.receive_data(data)
        except Exception as e:
            log.err()

    @property
    def session_id(self):
        return self.__session_id

class CommandListener(BaseProtocol):
    """The relay server will emit messages to us over a separate command 
    channel. The messages are referred to as commands, but are also referred-to 
    as "announcements".
    """    
    
    def __init__(self):
        self.__buffer = ReadBuffer()
    
    def connectionMade(self):
        pass

    def __handle_new_connection(self, properties):
        """We're about to receive data from a new client."""
        
        assigned_session_id = properties.assigned_to_session 
        
        log.msg("Received announcement of assignment to session-no "
                      "(%d)." % (assigned_session_id))
        
    def __handle_dropped_connection(self, properties):
        """The client assigned to us has dropped their connection. Ours will be
        dropped imminently."""
        
        session_id = properties.session_id

        log.msg("Received announcement of a connection drop for client "
                      "with session-no (%d)." % (session_id))

    def __handle_announcement(self, announcement):
        log.msg("Receiving an announcement with message-type of (%d)." % 
                      (announcement.message_type))
        
        if announcement.message_type == Command.CONNECTION_OPEN:
            self.__handle_new_connection(announcement.open_properties)
        elif announcement.message_type == Command.CONNECTION_DROP:
            self.__handle_dropped_connection(announcement.drop_properties)

    def dataReceived(self, data):
        log.msg("(%d) bytes of data received on command-channel." % (len(data)))
        
        try:
            self.__buffer.push(data)

            message_raw = self.__buffer.read_message()
            if message_raw is None:
                return

            log.msg("Message extracted from command-channel.")

            announcement = self.parse_or_raise(message_raw, Command)
            self.__handle_announcement(announcement)            
        except:
            log.err()


class HostProcessClientFactory(ReconnectingClientFactory):
    """This class manages instance-creation for the outgoing host-process 
    connections. It will reconnect if a connection is broken or times-out while 
    trying to connect.
    """

    def __repr__(self):
        return 'HostProcessClientFactory'
    
    def startedConnecting(self, connector):
        log.msg("Started to connect.")

    def buildProtocol(self, addr):
        log.msg("Connected host-process.")
        return HostProcess(self)


class CommandListenerClientFactory(ReconnectingClientFactory):
    """This class manages instance-creation for the outgoing command
    connections. It will reconnect if a connection is broken or times-out while 
    trying to connect.
    """

    def __repr__(self):
        return 'CommandListenerClientFactory'

    def startedConnecting(self, connector):
        log.msg("Started to connect.")

    def buildProtocol(self, addr):
        log.msg("Connected command-listener.")
        
        # We don't have any obligatory communication that happens at the top of
        # a command connection. Just mark it as successful, immediately.
        self.resetDelay()

        return CommandListener()

def main():
    parser = ArgumentParser(description="Start the host process and establish "
                                        "N connections to the relay server.")

    parser.add_argument('host', 
                        nargs='?', 
                        default='localhost', 
                        help="The hostname/IP of the relay-server.")
    
    parser.add_argument('-n', '--num-connections', 
                        nargs='?', 
                        default=10, 
                        type=int, 
                        help="Number of connections to maintain.")

    parser.add_argument('dport', 
                        nargs='?', 
                        default=8000, 
                        type=int, 
                        help="Port for host-process data-channel connections.")

    parser.add_argument('cport', 
                        nargs='?', 
                        default=8001, 
                        type=int, 
                        help="Port for host-process command-channel "
                             "connections.")    

    args = parser.parse_args()

    host = args.host
    num_connections = args.num_connections
    dport = args.dport
    cport = args.cport

    #startLogging(stdout)

    reactor.connectTCP(host, cport, CommandListenerClientFactory())

    # Spawn a series of connections to wait for incoming requests. As these are
    # "reconnecting" factories, they will all try to reconnect when their
    # connections are dropped after each client has finished-up (or for any 
    # other reason).
    i = num_connections
    while i > 0:
        reactor.connectTCP(host, dport, HostProcessClientFactory())
        i -= 1

    reactor.run()

if __name__ == '__main__':
    main()




from relayserver.message_types.command_pb2 import Command
from relayserver.message_types.hello_pb2 import HostProcessHelloResponse, Hello

def build_msg_cmd_connopen(assigned_hp_session):
        
    command = Command()
    command.version = 1
    command.message_type = Command.CONNECTION_OPEN
    command.open_properties.assigned_to_session = assigned_hp_session

    return command

def build_msg_cmd_conndrop(assigned_hp_session):
                
    command = Command()
    command.version = 1
    command.message_type = Command.CONNECTION_DROP
    command.drop_properties.session_id = assigned_hp_session
    
    return command

def build_msg_data_hphelloresponse(session_id, relay_host, relay_port):

    response = HostProcessHelloResponse()
    response.session_id = session_id
    response.relay_host = relay_host
    response.relay_port = str(relay_port)

    return response

def build_msg_data_hphello():
    hello = Hello()
    hello.version = 1

    return hello
